from __future__ import division

import random

import pygame

import data
import vector
import util

class Object(object):
    # aka mouse clickable
    selectable = False
    # an object is called with its position as the parameters
    def __init__(self, pos):
        self.pos = pos
    # dt is delta time. wtf?
    def update(self, dt, world):
        pass
    def draw(self, surf, level, selected):
        if level != self.level: return
        image = data.image[self.image_name]
        surf.blit(image, self.pos-vector.v(image.get_size())/2)

class Building(Object):
    pass

class Hub(Building):
    selectable = True
    image_name = "hub.png"
    level = 2
    def update(self, dt, world):
        Building.update(self, dt, world)
        self.string_cache = world.string
    def on_action(self, pos, world):
        if (pos - self.pos).mag() > world.string:
            pos = (pos-self.pos).norm()*world.string+self.pos
        if (pos - self.pos).mag() < 20: return
        if min((x.pos-self.pos).mag() for x in world.objects if isinstance(x, Spawn)) < 30: return
        world.string -= (pos - self.pos).mag()
        hub = Hub(pos)
        world.objects.append(Segment((self, hub)))
        world.objects.append(hub)
        world.selected = hub
    def draw(self, surf, level, selected):
        Building.draw(self, surf, level, selected)
        if selected and level == 4 and self.string_cache >= 20:
            pygame.draw.circle(surf, (255, 0, 0), self.pos, self.string_cache, 3)

class Spool(Hub):
    level = 2
    def draw(self, surf, level, selected):
        Hub.draw(self, surf, level, selected)
        if level == 0 or level == 2:
            image = data.image['spool.png']
            surf.blit(image, self.pos-vector.v(image.get_size())/2+(0, 10*(1-level)))

class Segment(Building):
    level = 1
    range = 40
    damage = 10
    recharge = 1 # secs between attk
    def __init__(self, hubs):
        assert len(hubs) == 2
        self.hubs = hubs
        self.recharge_counter = 0
    @property
    def pos(self):
        return self.hubs[0].pos/2 + self.hubs[1].pos/2
    def update(self, dt, world):
        self.recharge_counter -= dt
        if self.recharge_counter < 0:
            potential = []
            for object in world.objects:
                if not isinstance(object, Enemy): continue
                attk_pt = util.closest_point_on_line([x.pos for x in self.hubs], object.pos)
                dist = (object.pos-attk_pt).mag()
                #pygame.draw.circle(pygame.display.get_surface(), (0, 255, 0), attk_pt, 5)
                #pygame.display.update()
                if dist > self.range: continue
                potential.append(object)
            potential.sort(key=lambda x:x.health/x.damage)
            if potential:
                enemy = potential[0]
                attk_pt = util.closest_point_on_line([x.pos for x in self.hubs], enemy.pos)
                enemy.health_counter -= self.damage
                world.objects.append(Attack(attk_pt, enemy.pos))
                self.recharge_counter += self.recharge
        if self.recharge_counter < 0: self.recharge_counter = 0
    def draw(self, surf, level, selected):
        if level != self.level: return
        pygame.draw.aaline(surf, (255,255,255), self.hubs[0].pos, self.hubs[1].pos)

class Spawn(Object):
    image_name = "spawn.png"
    level = 0
    def __init__(self, *args):
        Object.__init__(self, *args)
        self.countdown = 5
    def update(self, dt, world):
        self.countdown -= dt
        while self.countdown < 0:
            world.objects.append(Enemy(self.pos))
            self.countdown += random.uniform(3, 6)

class Enemy(Object):
    image_name = 'enemy.png'
    level = 1
    health = 50
    damage = 5
    range = 20
    def __init__(self, *args):
        Object.__init__(self, *args)
        self.health_counter = self.health
    def update(self, dt, world):
        if self.health_counter <= 0:
            world.objects.remove(self)
            world.string += 40
            return
        dest = min((x for x in world.objects if isinstance(x, Hub)), key=lambda x: (x.pos-self.pos).mag())
        dir = dest.pos-self.pos
        dist = dir.mag()
        if dist > self.range:
            dir = dir / dist
            speed = min(20, self.range/dt)
            self.pos = self.pos + dir*speed*dt
            self.pos = self.pos + vector.v(-dir[1], dir[0])*random.uniform(-3, 3)*dt

class Effect(Object):
    level = 3
    def __init__(self, *args):
        Object.__init__(self, *args)
        self.timeleft = self.time
    def update(self, dt, world):
        self.timeleft -= dt
        if not self.timeleft > 0:
            world.objects.remove(self)

class Attack(Effect):
    time = .5
    def __init__(self, start, end):
        Effect.__init__(self, start/2+end/2)
        self.start = start
        self.end = end
    def draw(self, surf, level, selected):
        dir = (self.end-self.start).norm()
        perp = vector.v(-dir[1], dir[0])
        if level != self.level: return
        x = 0
        y = 0
        while x < (self.start-self.end).mag():
           ox = x
           oy = y
           x += random.uniform(3, 5)
           y = random.gauss(0, 3)
           pygame.draw.aaline(surf, (255, 255, 0), self.start+ox*dir+oy*perp, self.start+x*dir+y*perp)

class World(object):
    def __init__(self, size):
        self.size = size
        self.objects = list()
        self.objects.append(Spool(self.size*vector.random(2)))
        self.objects.append(Spawn(self.size*vector.random(2)))
        self.selected = self.objects[0]
        self.string = 300
    def on_click(self, pos, button):
        if button == 1: # select
            if not self.objects: return
            self.selected = min((x for x in self.objects if x.selectable), key=lambda x:(x.pos-pos).mag())
        elif button == 3:
            if not self.selected: return
            self.selected.on_action(pos, self)
    def update(self, dt):
        for object in list(self.objects): # copied so objects can delete themselves
            object.update(dt, self)
    def draw(self, surf):
        """
        levels:
            ground
            0 - things flat on ground
            1 - things above ground
            2 - highest ground things
            3 - effects
            4 - selection
        """
        self.objects.sort(key=lambda x: x.pos[1])
        for level in xrange(5):
            for object in self.objects:
                object.draw(surf, level, object is self.selected)
